"""
Input fields

input_id, input_address, flap_score, flap_uprn, label_uprn optionals ['uprn_row', 'error']

Output

Evaluation metrics

Per strata-value, micro metrics

Per strata, macro metrics

Average of per strata, final metrics


The process for eval_by_sampling_for_val

<input> = ['input_id', 'input_address', OPTIONAL: 'uprn']
<sampling> = ['strata', 'strata_value', 'sample_weight']
<verdict> = ['linking_verdict', 'clf_verdict']
<metrics> = ['accuracy', 'accuracy_strict', 'precision', 'recall', 'f1']
<linking_verdict> = ['FN-c_ps_npr', 'FP_Not_Possible-nc_nps_pr',
       'FP_Wrong_One-nc_ps_pr', 'Good-TN-nc_nps_npr', 'Good-TP-c_ps_pr',
       'TN-Not_found-nc_ps_npr']
<clf_verdict> = ['clf_FN', 'clf_FP', 'clf_TN', 'clf_TP']

- Sampling the input addresses
    - Match and Parce: data[<input>]
        --> data[<input>, <matching>, <parsing>]
    - Sample:
        --> sampled[<input>, <matching>, <sampling>]

- Review and Add true UPRN
    - sampled[<input>, <matching>, <sampling>] --> sampled_with_review[..., 'review']
    review should take 3 kinds of values
    'y' is correct
    'n' is not possible
    regex'\\d+' or numeric string to provide the correct UPRN

- Summarise
    - Add verdict
        sampled_with_review[<input>, <matching>, <sampling>]
        --> sampled_with_review[..., 'linking_verdict', 'clf_verdict']
    - Summarise performance
        sampled_with_review --> performance['name_of_group', <linking_verdicts>, <clf_verdicts>, <metrics>]


Mode `val` validation
Sampling --> Manual Review --> Summarise

Mode `dev` development
Sampling --> Add Review --> Summarise

"""

import warnings
import re
import os

import pandas as pd
from flap.parser.rule_parser_fast import RuleParserFast
from flap.database.sql import SqlDB
from flap.api.score_and_match import score_and_match


def length_number_like(row):
    return sum([row['number_like_%s' % i] != '' for i in range(5)])


def prepare_input_for_sampling(input_csv, db_path, spd_path=None, classifier_model_path=None,
                               random_sample_size=10000, random_state=666):

    if isinstance(input_csv, str):
        data = pd.read_csv(input_csv, dtype='object', usecols=['input_id', 'input_address', 'uprn']).sample(
            random_sample_size, random_state=random_state)

    elif isinstance(input_csv, pd.DataFrame):
        data = input_csv[['input_id', 'input_address', 'uprn']].sample(
            random_sample_size, random_state=random_state).copy()
    else:
        raise TypeError('input_csv should be path(str) or pd.DataFrame')

    # Match

    if 'flap_match_score' not in data:
        match_results = score_and_match(data, db_path, classifier_model_path=classifier_model_path,
                                        score_batch_size=random_sample_size,
                                        match_batch_size=int(random_sample_size/20))

        data = pd.merge(left=data, right=match_results[['input_id', 'flap_match_score', 'flap_uprn', 'uprn_row']],
                        on='input_id', how='left')

    data['flap_score_decile'] = (data.flap_match_score.astype(float) * 10 - 1e-8).astype(int)

    # Parse
    sql_db = SqlDB(db_path)

    parser = RuleParserFast(sql_db)

    if 'number_like_0' not in data:
        data = parser.parse_batch_in_df(data)

    p_multi = re.compile(r'\d[A-Z]\d+')
    data['has_multiplier'] = data.input_address.apply(lambda x: re.search(p_multi, x) is not None)

    data['length_number_like_parsing'] = data.apply(length_number_like, axis=1)

    if spd_path is not None:

        spd = pd.read_csv(spd_path, dtype='object',
                          usecols=['Postcode', 'UrbanRural8Fold2020Code', 'ScottishIndexOfMultipleDeprivation2020Rank'])

        spd_name_mapper = {
            'Postcode': 'postcode_spd',
            'UrbanRural8Fold2020Code': 'ur_8fold',
            'ScottishIndexOfMultipleDeprivation2020Rank': 'simd_2020'
        }

        spd = spd.rename(mapper=spd_name_mapper, axis=1)

        spd = spd[~spd.postcode_spd.duplicated()]

        data = pd.merge(left=data, right=spd, left_on='POSTCODE', right_on='postcode_spd', how='left')

    return data


def stratified_sampling_by_column_value(data, **stratify_config):
    sampled_groups = []

    for col in stratify_config:
        try:
            for i, group in data.groupby(col, dropna=False):
                n = min(stratify_config[col], len(group))

                sampled_group = group[['input_id', 'input_address', 'uprn', 'uprn_row',
                                       'flap_match_score', 'flap_uprn']].sample(n)
                sampled_group['strata'] = col
                sampled_group['strata_value'] = i
                sampled_group['sample_weight'] = len(group) / len(data) / n
                sampled_group['review'] = ''

                sampled_groups.append(sampled_group)

        except KeyError:

            warnings.warn('Columns %s does not exist, skipped for sampling' % col)

    sampled = pd.concat(sampled_groups)

    return sampled


def stratified_sampling_for_review(input_csv, db_path, spd_path=None,
                                   random_sample_size=10000, random_state=666, **stratify_config):
    data = prepare_input_for_sampling(input_csv=input_csv, db_path=db_path, spd_path=spd_path,
                                      random_sample_size=random_sample_size, random_state=random_state)

    sampled = stratified_sampling_by_column_value(data, **stratify_config)

    return sampled


def uprn_in_db(uprn, sql_db):
    query = 'select UPRN from indexed where UPRN=="%s"' % uprn

    return len(sql_db.sql_query(query=query)) > 0


def add_review(sampled, sql_db):
    sampled['review'] = sampled.apply(
        lambda row: 'y' if row['uprn'] == row['flap_uprn'] else 'n'
        if not uprn_in_db(row['uprn'], sql_db) else row['uprn'], axis=1)

    sampled_with_review = sampled

    return sampled_with_review


def _add_linking_verdict_to_reviewed_samples(row):
    correct = True if row['review'] == 'y' else False
    possible = False if row['review'] == 'n' else True
    y_pred = True if row['flap_match_score'] > 0.5 else False

    if correct and possible and y_pred:
        return 'Good-TP-c_ps_pr'

    if correct and possible and not y_pred:
        return 'FN-c_ps_npr'

    if not correct and possible and y_pred:
        return 'FP_Wrong_One-nc_ps_pr'

    if not correct and not possible and y_pred:
        return 'FP_Not_Possible-nc_nps_pr'

    if not correct and possible and not y_pred:
        return 'TN-Not_found-nc_ps_npr'

    if not correct and not possible and not y_pred:
        return 'Good-TN-nc_nps_npr'


def _add_classifier_verdict_to_reviewed_samples(row):
    correct = True if row['review'] == 'y' else False
    y_pred = True if row['flap_match_score'] > 0.5 else False

    if correct and y_pred:
        return 'clf_TP'

    if correct and not y_pred:
        return 'clf_FN'

    if not correct and y_pred:
        return 'clf_FP'

    if not correct and not y_pred:
        return 'clf_TN'


def add_verdict(sampled_with_review):
    sampled_with_review['link_verdict'] = sampled_with_review.apply(_add_linking_verdict_to_reviewed_samples, axis=1)
    sampled_with_review['clf_verdict'] = sampled_with_review.apply(_add_classifier_verdict_to_reviewed_samples, axis=1)

    return sampled_with_review


def summarise_performance_from_verdict(group, add_name=None):
    record = {
        'name_of_group': add_name,
        'FN-c_ps_npr': 0,
        'FP_Not_Possible-nc_nps_pr': 0,
        'FP_Wrong_One-nc_ps_pr': 0,
        'Good-TN-nc_nps_npr': 0,
        'Good-TP-c_ps_pr': 0,
        'TN-Not_found-nc_ps_npr': 0,
        'clf_FN': 0,
        'clf_FP': 0,
        'clf_TN': 0,
        'clf_TP': 0,
        'accuracy': 0,
        'accuracy_strict': 0,
        'precision': 0,
        'recall': 0,
        'f1': 0
    }

    record.update(group.groupby('link_verdict')['sample_weight'].sum().to_dict())
    record.update(group.groupby('clf_verdict')['sample_weight'].sum().to_dict())

    record['accuracy'] = record['FN-c_ps_npr'] + record['Good-TN-nc_nps_npr'] + record['Good-TP-c_ps_pr']
    record['accuracy_strict'] = record['Good-TN-nc_nps_npr'] + record['Good-TP-c_ps_pr']

    try:
        record['precision'] = record['clf_TP'] / (record['clf_TP'] + record['clf_FP'])
        record['recall'] = record['clf_TP'] / (record['clf_TP'] + record['clf_FN'])
        record['f1'] = 2 * record['clf_TP'] / (2 * record['clf_TP'] + record['clf_FP'] + + record['clf_FN'])
    except ZeroDivisionError:
        pass

    return record


def summarise_performance_in_stratified_sampled(sampled_for_review):
    records = []

    for i, group in sampled_for_review.groupby('strata'):
        records.append(summarise_performance_from_verdict(group, add_name=i))

    performance = pd.DataFrame.from_records(records)

    average_row = performance.iloc[:, 1:].apply(lambda x: sum(x) / len(x)).to_dict()

    average_row['name_of_group'] = 'average'

    records.append(average_row)

    for i, group in sampled_for_review.groupby('strata'):

        if group.strata_value.nunique() <= 10:

            for ii, gg in group.groupby('strata_value'):
                records.append(summarise_performance_from_verdict(gg, add_name='%s-%s' % (i, ii)))

    performance = pd.DataFrame.from_records(records)

    return performance


def eval_for_dev(input_csv, db_path, save_sampled=None, save_performance=None,
                 spd_path=None, random_sample_size=10000, random_state=666, **stratify_config):
    data = prepare_input_for_sampling(input_csv=input_csv, db_path=db_path, spd_path=spd_path,
                                      random_sample_size=random_sample_size, random_state=random_state)

    sampled = stratified_sampling_by_column_value(data, **stratify_config)

    if save_sampled is not None:
        sampled.to_csv(save_sampled)

    sql_db = SqlDB(db_path)
    sampled_with_review = add_review(sampled, sql_db)

    sampled_with_review = add_verdict(sampled_with_review)

    performance = summarise_performance_in_stratified_sampled(sampled_with_review)

    if save_performance is not None:
        performance.to_csv(save_performance)

    return performance



import pandas as pd
from tqdm import tqdm
from flap.api.score import score
from flap.api.match import match


def score_and_match(input_csv, db_path, final_output=None,
                    max_workers=None, in_memory_db=False, classifier_model_path=None,
                    input_address_col='input_address', uprn_col='uprn',
                    score_output=None, score_output_raw=None, score_batch_size=100000,
                    match_output=None, match_output_raw=None, match_batch_size=10000,
                    max_beam_width=200, score_threshold=0.3):

    scored = score(input_csv, db_path, output_file_path=score_output, raw_output_path=score_output_raw,
                   batch_size=score_batch_size, max_workers=max_workers, in_memory_db=in_memory_db,
                   classifier_model_path=classifier_model_path,
                   input_address_col=input_address_col, uprn_col=uprn_col)

    scored['flap_eval_score'] = scored['flap_eval_score'].astype(float)

    scored_to_flap = scored[(scored.flap_eval_score < 0.5) | scored.flap_eval_score.isna()]

    scored_skipped = scored[scored.flap_eval_score >= 0.5].copy()
    scored_skipped['flap_match_score'] = scored_skipped['flap_eval_score']
    scored_skipped['flap_uprn'] = scored_skipped['uprn']

    matched = match(input_csv=scored_to_flap, db_path=db_path, output_file_path=match_output,
                    raw_output_path=match_output_raw,
                    batch_size=match_batch_size, max_workers=max_workers, in_memory_db=in_memory_db,
                    classifier_model_path=classifier_model_path,
                    max_beam_width=max_beam_width, score_threshold=score_threshold
                    )

    final = pd.concat([scored_skipped, matched])

    if final_output is not None:
        final.to_csv(final_output)

    return final

# input_csv = '/home/huayu_ssh/PycharmProjects/dres_r/projects/phs_tests/sampled_ctax_matched_proc.csv'
# db_path = '/home/huayu_ssh/PycharmProjects/dres_r/db/test_db'
#
# max_workers = None
#
# in_memory_db = False
# classifier_model_path=None
#
# input_address_col='input_address'
#
# final_output = '/home/huayu_ssh/PycharmProjects/dres_r/projects/phs_tests/final.csv'
#
# # Score args
# score_output = None
# score_output_raw = '/home/huayu_ssh/PycharmProjects/dres_r/projects/phs_tests/scoring_raw_output'
# score_batch_size = 100000
#
# uprn_col='uprn'
#
# # Match args
#
# match_output = None
# match_output_raw = '/home/huayu_ssh/PycharmProjects/dres_r/projects/phs_tests/match_raw_output'
#
# match_batch_size = 10000
# max_beam_width=200
# score_threshold=0.3
#



"""
Top level API for scoring existing matches

"""

import pandas as pd
import numpy as np
import csv
import os

import traceback

from flap.database.sql import SqlDB
from flap.matcher.sql_matcher import SqlMatcher
from flap.database.sql_in_memory import SqlDBInMemory
from flap.parser.rule_parser_fast import RuleParserFast
from flap.utils import join_uprn_fields, available_cpu_count


def csv_row_counter(filename):

    with open(filename, 'r') as f:
        row_count = sum(1 for _ in csv.reader(f))

    return row_count


def read_csv_header(filename):

    with open(filename, 'r') as f:
        dict_reader = csv.DictReader(f)
        headers = dict_reader.fieldnames

    return headers


def _load_all_csv_from_path(path, **kwargs):
    csv_files = [os.path.join(path, p) for p in os.listdir(path) if '.csv' in p]

    index_exists = _check_index(csv_files[0])

    if index_exists:
        dfs = [pd.read_csv(p, index_col=0, **kwargs) for p in csv_files]
    else:
        dfs = [pd.read_csv(p, **kwargs) for p in csv_files]

    return pd.concat(dfs)


def _check_index(csv_file):
    with open(csv_file, 'r') as f:
        header_0 = f.readline().split(',')[0]

    return header_0 == ''


def generate_chunks_of_dataframe(df, chunk_size):

    i = 0

    while (i + chunk_size) < len(df):

        yield df.iloc[i:(i + chunk_size)]

        i += chunk_size

    yield df.iloc[i:]


def score(input_csv, db_path, output_file_path=None, raw_output_path=None,
          batch_size=10000, max_workers=None, in_memory_db=False, classifier_model_path=None,
          input_address_col='input_address', uprn_col='uprn'):

    # Initialise parameters

    # if output_file_path is None:
    #     output_file_path = os.path.join(os.getcwd(), 'scoring_output.csv')

    if raw_output_path is not None:

        raw_output_path = os.path.join(os.getcwd(), 'scoring_raw_output')

        if not os.path.exists(raw_output_path):
            os.mkdir(raw_output_path)

    if max_workers is None:
        max_workers = available_cpu_count()

    # Check and read the input

    if isinstance(input_csv, str):
        batch_size_adj = int(batch_size / max_workers) * max_workers
        total_tasks = csv_row_counter(input_csv) - 1
        total_batches = int(total_tasks / batch_size_adj) + int((total_tasks % batch_size_adj) > 0)

        headers = read_csv_header(input_csv)
        assert all(s in headers for s in ['input_id', 'input_address']), \
            'Two columns are required in input csv file: `input_id` and `input_address`'
        batch_gen = pd.read_csv(input_csv, dtype='object', chunksize=batch_size_adj, index_col=0)

    elif isinstance(input_csv, pd.DataFrame):
        batch_size_adj = int(batch_size / max_workers) * max_workers
        total_tasks = len(input_csv)
        total_batches = int(total_tasks / batch_size_adj) + int((total_tasks % batch_size_adj) > 0)

        headers = input_csv.columns
        assert all(s in headers for s in ['input_id', 'input_address']), \
            'Two columns are required in input csv file: `input_id` and `input_address`'
        batch_gen = generate_chunks_of_dataframe(input_csv, chunk_size=batch_size_adj)

    else:
        raise TypeError('input_csv should be path(str) or pd.DataFrame')

    # Check the database
    if not in_memory_db:
        sql_db = SqlDB(db_path)
        assert sql_db.db_status['table_indexed_built'], 'Database is not indexed, please Build the Database first'

        matcher = SqlMatcher(sql_db, scorer_path=classifier_model_path)
    else:
        sql_db = SqlDB(db_path)

        sql_db_in_memory = SqlDBInMemory()

        csv_files = [os.path.join(sql_db.sub_paths['csv'], file) for file in os.listdir(sql_db.sub_paths['csv'])]
        csv_names = [os.path.basename(file).split('.')[0]
                     for file in os.listdir(sql_db.sub_paths['csv'])]

        assert all([s in csv_names for s in ['indexed', 'expanded']]), 'CSV database does not exist'

        for table_name, path in zip(csv_names, csv_files):
            print(f'Loading Table {table_name} in memory')
            sql_db_in_memory.load_csv(path, table_name)
            print()

        parser = RuleParserFast(sql_db)

        matcher = SqlMatcher(sql_db_in_memory, parser=parser, scorer_path=classifier_model_path)

    # Main scoring loop

    batch_index = 0

    res_collection = []

    while True:

        try:
            batch_name = 'batch_%s.csv' % batch_index
            print('Processing %s out of %s' % (batch_name, total_batches))

            batch_exists = False

            if raw_output_path is not None:
                batch_path = os.path.join(raw_output_path, batch_name)
                batch_exists = os.path.exists(batch_path)

            df_batch = next(batch_gen).copy()

            if (not batch_exists) and len(df_batch):

                mapper = {input_address_col: 'input_address', uprn_col: 'uprn'}
                rev_mapper = {'input_address': input_address_col, 'uprn': uprn_col}
                df_batch.rename(mapper, axis='columns', inplace=True)

                chunk_size = int(batch_size / max_workers) + 1

                res = matcher.score_matching_of_batch(df_batch, input_address_col='input_address', uprn_col='uprn',
                                                      max_workers=max_workers, chunksize=chunk_size)

                res.rename(rev_mapper, axis='columns', inplace=True)

                if raw_output_path is not None:
                    batch_path = os.path.join(raw_output_path, batch_name)
                    res.to_csv(batch_path)
                else:
                    res_collection.append(res.copy())

            batch_index += 1

        except StopIteration:
            print('Scoring Finished, start summarising results')
            break

    if raw_output_path is not None:
        results = _load_all_csv_from_path(raw_output_path, dtype='object')
    else:
        results = pd.concat(res_collection)

    if output_file_path is not None:
        results.to_csv(output_file_path)
        print('Results can be see at: %s' % output_file_path)

    return results



"""
Top level API for address matching
"""

import csv
import os

import pandas as pd
import traceback

from flap.database.sql import SqlDB
from flap.database.sql_in_memory import SqlDBInMemory
from flap.matcher.sql_matcher import SqlMatcher
from flap.parser.rule_parser_fast import RuleParserFast
from flap.utils import join_uprn_fields, available_cpu_count


def csv_row_counter(filename):

    with open(filename, 'r') as f:
        row_count = sum(1 for _ in csv.reader(f))

    return row_count


def read_csv_header(filename):
    with open(filename, 'r') as f:
        dict_reader = csv.DictReader(f)
        headers = dict_reader.fieldnames

    return headers


def _load_all_csv_from_path(path, **kwargs):
    csv_files = [os.path.join(path, p) for p in os.listdir(path) if '.csv' in p]

    index_exists = _check_index(csv_files[0])

    if index_exists:
        dfs = [pd.read_csv(p, index_col=0, **kwargs) for p in csv_files]
    else:
        dfs = [pd.read_csv(p, **kwargs) for p in csv_files]

    return pd.concat(dfs)


def _check_index(csv_file):
    with open(csv_file, 'r') as f:
        header_0 = f.readline().split(',')[0]

    return header_0 == ''


def generate_chunks_of_dataframe(df, chunk_size):

    i = 0

    while (i + chunk_size) < len(df):

        yield df.iloc[i:(i + chunk_size)]

        i += chunk_size

    yield df.iloc[i:]


def match(input_csv, db_path, output_file_path=None, raw_output_path=None,
          batch_size=10000, max_workers=None, in_memory_db=False, classifier_model_path=None,
          max_beam_width=200, score_threshold=0.3):
    """
    This is the top level function for matching free-text addresses from a csv file to the OS ABP UPRN database.

    Parameters
    ----------
    input_csv : str
        Path to the csv file. The file needs to have two fields in the header ['input_id', 'input_address']
    db_path : str
        Path to the database built. See `flap.create_db()`
    output_file_path : str, default None
        Path for saving the output csv file, containing ['input_id', 'input_address', 'uprn', 'score']. If None, results
        is saved to '[$pwd]/output.csv'
    raw_output_path : str, default None
        Path for save the batched raw output files. If None, raw output is saved to '[$pwd]/output_raw/'
    batch_size : int, default 10000
        Size of each batch
    max_workers : int, default None
        Number of processes. If None, the max cpu available is determined by `flap.utils.cpu_count.available_cpu_count()`
    in_memory_db : bool, default False
        If in-memory SQLite database is used. If True, a temp database is created in shared memory cache from pre-built
        csv files
    classifier_model_path : str, default None
        The path to the pretrained sklearn classifier model. If None, the model is loaded from 'flap.__file__/*.clf'
    max_beam_width: int, default 200
        The max number of rows to be considered from UPRN database
    score_threshold: float, default 0.3
        The min score for early stop of matching
    Returns
    -------
    pandas.DataFrame
        Match results
    Examples
    --------
    >>> from flap import match
    >>> input_csv = '...'
    >>> db_path = '...'
    >>> results = match(
    ...    input_csv=input_csv,
    ...    db_path=db_path
    ...)
    >>> print(results)
    """
    # Initialise parameters

    # if output_file_path is None:
    #     output_file_path = os.path.join(os.getcwd(), 'matching_output.csv')

    if raw_output_path is not None:

        raw_output_path = os.path.join(os.getcwd(), 'scoring_raw_output')

        if not os.path.exists(raw_output_path):
            os.mkdir(raw_output_path)

    if max_workers is None:
        max_workers = available_cpu_count()

    # Check and read the input

    if isinstance(input_csv, str):
        batch_size_adj = int(batch_size / max_workers) * max_workers
        total_tasks = csv_row_counter(input_csv) - 1
        total_batches = int(total_tasks / batch_size_adj) + int((total_tasks % batch_size_adj) > 0)

        headers = read_csv_header(input_csv)
        assert all(s in headers for s in ['input_id', 'input_address']), \
            'Two columns are required in input csv file: `input_id` and `input_address`'
        batch_gen = pd.read_csv(input_csv, dtype='object', chunksize=batch_size_adj, index_col=0)

    elif isinstance(input_csv, pd.DataFrame):
        batch_size_adj = int(batch_size / max_workers) * max_workers
        total_tasks = len(input_csv)
        total_batches = int(total_tasks / batch_size_adj) + int((total_tasks % batch_size_adj) > 0)

        headers = input_csv.columns
        assert all(s in headers for s in ['input_id', 'input_address']), \
            'Two columns are required in input csv file: `input_id` and `input_address`'
        batch_gen = generate_chunks_of_dataframe(input_csv, chunk_size=batch_size_adj)

    else:
        raise TypeError('input_csv should be path(str) or pd.DataFrame')

    # Check the database
    if not in_memory_db:
        sql_db = SqlDB(db_path)
        assert sql_db.db_status['table_indexed_built'], 'Database is not indexed, please Build the Database first'

        matcher = SqlMatcher(sql_db, scorer_path=classifier_model_path)
    else:
        sql_db = SqlDB(db_path)

        sql_db_in_memory = SqlDBInMemory()

        csv_files = [os.path.join(sql_db.sub_paths['csv'], file) for file in os.listdir(sql_db.sub_paths['csv'])]
        csv_names = [os.path.basename(file).split('.')[0]
                     for file in os.listdir(sql_db.sub_paths['csv'])]

        assert all([s in csv_names for s in ['indexed', 'expanded']]), 'CSV database does not exist'

        for table_name, path in zip(csv_names, csv_files):

            print(f'Loading Table {table_name} in memory')
            sql_db_in_memory.load_csv(path, table_name)
            print()

        parser = RuleParserFast(sql_db)

        matcher = SqlMatcher(sql_db_in_memory, parser=parser, scorer_path=classifier_model_path)

    matcher.set_match_config(max_beam_width=max_beam_width, score_threshold=score_threshold)

    # Main matching loop

    batch_index = 0

    res_collection = []

    while True:

        try:
            batch_name = 'batch_%s.csv' % batch_index
            print('Processing %s out of %s' % (batch_name, total_batches))


            batch_exists = False

            if raw_output_path is not None:
                batch_path = os.path.join(raw_output_path, batch_name)
                batch_exists = os.path.exists(batch_path)

            df_batch = next(batch_gen).copy()

            if (not batch_exists) and len(df_batch):

                address_list = df_batch.input_address.to_list()

                chunk_size = int(batch_size_adj / max_workers / 4) + int((batch_size_adj % max_workers) > 0)

                results = matcher.match_batch(address_list, max_workers=max_workers, chunksize=chunk_size)

                records = []

                for (_, row), result in zip(df_batch.iterrows(), results):

                    record = {
                        'input_id': row['input_id'],
                        'input_address': row['input_address'],
                    }

                    try:
                        record['uprn_row'] = join_uprn_fields(result['uprn_row'])
                        record['flap_uprn'] = result['uprn_row']['UPRN']
                        record['flap_match_score'] = result['score']

                    except KeyError:
                        try:
                            record['error'] = result['error']
                        except:
                            record['error'] = traceback.format_exc()

                    except TypeError:
                        record['error'] = traceback.format_exc()

                    records.append(record)

                records_df = pd.DataFrame.from_records(records)

                if raw_output_path is not None:
                    batch_path = os.path.join(raw_output_path, batch_name)
                    records_df.to_csv(batch_path)
                else:
                    res_collection.append(records_df.copy())

            batch_index += 1

        except StopIteration:
            print('Matching Finished, start summarising results')
            break

    # Summarising results

    if raw_output_path is not None:
        results = _load_all_csv_from_path(raw_output_path, dtype='object')
    else:
        results = pd.concat(res_collection)

    if output_file_path is not None:
        results.to_csv(output_file_path)
        print('Results can be see at: %s' % output_file_path)

    return results



import pandas as pd
import json
import os
import sqlite3
import shutil
from warnings import warn
from tqdm import tqdm

import flap

from flap.database.db_import import db_import
from flap.database.db_build_vocabulary import compile_pc1_uniques, compile_pc1_mappings_region, compile_global_uniques
from flap.database.db_index import db_index


MODULE_PATH = os.path.dirname(flap.__file__)


class SqlDBManager:

    def __init__(self, project_name=None):
        self.global_db_path = os.path.join(MODULE_PATH, 'db')
        self.project_name = project_name
        if project_name is not None:
            self.project_db_path = os.path.join(os.getcwd(), project_name, 'db')

    def list_global_db(self):
        db_names = os.listdir(self.global_db_path)
        return db_names

    def list_project_db(self):
        db_names = os.listdir(self.project_db_path)
        return db_names

    @staticmethod
    def create_db(path):
        new_db_path = path

        db_name = os.path.basename(path)

        if os.path.exists(new_db_path):
            warn('DB:%s exists, new database cannot be created.' % db_name)
        else:
            os.makedirs(new_db_path, exist_ok=True)
            print('New db created at: %s' % new_db_path)
            os.mkdir(os.path.join(new_db_path, 'raw'))
            os.mkdir(os.path.join(new_db_path, 'sql'))
            os.mkdir(os.path.join(new_db_path, 'db_config'))
            os.mkdir(os.path.join(new_db_path, 'csv'))

    def create_global_db(self, db_name):
        new_db_path = os.path.join(self.global_db_path, db_name)

        if os.path.exists(new_db_path):
            warn('DB:%s exists, new database cannot be created.' % db_name)
        else:
            os.makedirs(new_db_path, exist_ok=True)
            print('New db created at: %s' % new_db_path)
            os.mkdir(os.path.join(new_db_path, 'raw'))
            os.mkdir(os.path.join(new_db_path, 'sql'))
            os.mkdir(os.path.join(new_db_path, 'db_config'))
            os.mkdir(os.path.join(new_db_path, 'csv'))

    def create_project_db(self, db_name):
        new_db_path = os.path.join(self.project_db_path, db_name)

        if os.path.exists(new_db_path):
            warn('DB:%s exists, new database cannot be created.' % db_name)
        else:
            os.makedirs(new_db_path, exist_ok=True)
            print('New db created at: %s' % new_db_path)
            os.mkdir(os.path.join(new_db_path, 'raw'))
            os.mkdir(os.path.join(new_db_path, 'sql'))
            os.mkdir(os.path.join(new_db_path, 'db_config'))
            os.mkdir(os.path.join(new_db_path, 'csv'))

    def get_db_path(self, db_name, project_level=False):
        if not project_level:
            if db_name in self.list_global_db():
                return os.path.join(self.global_db_path, db_name)
            else:
                warn('DB: %s does not exist!' % db_name)
        else:
            if self.project_name is None:
                warn('Project not specified')
            else:
                if db_name in self.list_project_db():
                    return os.path.join(self.project_db_path, db_name)
                else:
                    warn('DB: %s does not exist!' % db_name)

    def get_db(self, db_name, project_level=False):

        return SqlDB(self.get_db_path(db_name, project_level))


class SqlDB:

    def __init__(self, path_to_db):
        self.db_path = path_to_db
        self.db_name = os.path.basename(self.db_path)
        self.sub_paths = {
            'raw': os.path.join(self.db_path, 'raw'),
            'db_config': os.path.join(self.db_path, 'db_config'),
            'sql_db': os.path.join(self.db_path, 'sql', 'db.sqlite3'),
            'sql_db_temp': os.path.join(self.db_path, 'sql', 'db_temp.sqlite3'),
            'csv': os.path.join(self.db_path, 'csv'),
            'vocabulary': os.path.join(self.db_path, 'vocabulary'),
            'pc1_mappings': os.path.join(self.db_path, 'vocabulary', 'pc1_mappings.json'),
            'pc1_mappings_region': os.path.join(self.db_path, 'vocabulary', 'pc1_mappings_region.json'),
            'unique_DOUBLE_DEPENDENT_LOCALITY': os.path.join(self.db_path, 'vocabulary',
                                                             'unique_DOUBLE_DEPENDENT_LOCALITY.json'),
            'unique_DEPENDENT_LOCALITY': os.path.join(self.db_path, 'vocabulary', 'unique_DEPENDENT_LOCALITY.json'),
            'unique_POST_TOWN': os.path.join(self.db_path, 'vocabulary', 'unique_POST_TOWN.json'),
            'thoroughfare_patterns': os.path.join(self.db_path, 'vocabulary', 'thoroughfare_patterns.json')
        }

        self.__db_status = None

    def __str__(self):
        self_str = f"""
        Database name: {self.db_name}
        Database path: {self.db_path}
        """
        return self_str

    def get_table_names(self):

        conn = self.get_conn()
        cur = conn.cursor()

        try:
            cur.execute("""SELECT name FROM sqlite_schema""")
            res = cur.fetchall()

            if len(res):
                table_names = [list(name)[0] for name in res]
            else:
                table_names = []

        except sqlite3.OperationalError:
            table_names = []

        cur.close()
        conn.close()

        return table_names

    def get_columns_of_table(self, table_name):

        table_name = (table_name,)

        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute("""
        SELECT sql FROM sqlite_master WHERE tbl_name = ? AND type = 'table'
        """, table_name)
        res = cur.fetchall()

        try:
            s = res[0][0]
            lines = s.split('\n')
            column_names = [line.split('"')[1] for line in lines[1:-1]]
        except TypeError:
            column_names = []
        except IndexError:
            column_names = []

        cur.close()
        conn.close()

        return column_names

    @property
    def db_status(self):
        db_status = {
            'db_name': self.db_name,
            'db_path': self.db_path,
            'valid': os.path.isdir(self.db_path),
            'raw_added': len(os.listdir(self.sub_paths['raw'])) > 0
        }

        tables = self.get_table_names()

        db_status['table_raw_built'] = 'raw' in tables
        db_status['table_indexed_built'] = 'indexed' in tables

        self.__db_status = db_status

        return self.__db_status

    def drop_table(self, table):
        conn = self.get_conn()
        cur = conn.cursor()

        try:
            cur.execute(f"""DROP TABLE {table}""")

        except sqlite3.OperationalError:
            print('Table not exist')

        conn.commit()
        cur.close()
        conn.close()

    def get_db_config(self):

        sql_config_path = [os.path.join(self.sub_paths['db_config'], file)
                           for file in os.listdir(self.sub_paths['db_config']) if '.json' in file]

        if len(sql_config_path):

            sql_config_file = sql_config_path[0]

            with open(sql_config_file, 'r') as f:
                sql_config = json.load(f)

        else:

            sql_config = {
                "RECORD_IDENTIFIER": "INTEGER",
                "CHANGE_TYPE": "TEXT",
                "PRO_ORDER": "TEXT",
                "UPRN": "TEXT",
                "UDPRN": "TEXT",
                "ORGANISATION_NAME": "TEXT",
                "DEPARTMENT_NAME": "TEXT",
                "SUB_BUILDING_NAME": "TEXT",
                "BUILDING_NAME": "TEXT",
                "BUILDING_NUMBER": "TEXT",
                "DEPENDENT_THOROUGHFARE": "TEXT",
                "THOROUGHFARE": "TEXT",
                "DOUBLE_DEPENDENT_LOCALITY": "TEXT",
                "DEPENDENT_LOCALITY": "TEXT",
                "POST_TOWN": "TEXT",
                "POSTCODE": "TEXT",
                "POSTCODE_TYPE": "TEXT",
                "DELIVERY_POINT_SUFFIX": "TEXT",
                "WELSH_DEPENDENT_THOROUGHFARE": "TEXT",
                "WELSH_THOROUGHFARE": "TEXT",
                "WELSH_DOUBLE_DEPENDENT_LOCALITY": "TEXT",
                "WELSH_DEPENDENT_LOCALITY": "TEXT",
                "WELSH_POST_TOWN": "TEXT",
                "PO_BOX_NUMBER": "TEXT",
                "PROCESS_DATE": "TEXT",
                "START_DATE": "TEXT",
                "END_DATE": "TEXT",
                "LAST_UPDATE_DATE": "TEXT",
                "ENTRY_DATE": "TEXT"
            }

            sql_config_file = os.path.join(self.sub_paths['db_config'], 'db_config.json')

            with open(sql_config_file, 'w') as f:
                json.dump(sql_config, f, indent=4)

            print('db_config created')
        return sql_config

    def setup_database(self, if_exists='skip'):
        self.build_raw(if_exists=if_exists)
        self.build_vocabulary()
        self.indexing_db(if_exists=if_exists)
        self.build_csv()
        print('Database Setup complete!')

    def build(self, if_exists='skip'):
        self.build_raw(if_exists=if_exists)
        self.build_vocabulary()
        self.indexing_db(if_exists=if_exists)
        self.build_csv()
        print('Database Setup complete!')
        
    def build_raw(self, if_exists='skip'):
        db_status = self.db_status

        print(db_status)

        if not db_status['raw_added']:
            raise FileNotFoundError('Add raw database files to [db_path]/raw first and then run build')

        if if_exists == 'skip':

            if db_status['table_raw_built']:
                print('Table raw exists, build skipped. Use if_exists="replace", if you want to create a new db')
            else:
                db_import(self)

        elif if_exists == 'replace':

            self.drop_table('raw')
            db_import(self)

    def indexing_db(self, if_exists='skip'):
        db_status = self.db_status

        if if_exists == 'skip':

            if db_status['table_indexed_built']:
                print('TABLE indexed exists, build skipped. Use if_exists="replace", if you want to create a new db')
            else:
                db_index(self)
                self.delete_temp()

        elif if_exists == 'replace':

            self.drop_table('indexed')
            db_index(self)
            self.delete_temp()

    def build_vocabulary(self):

        print('Start building vocabularies')
        if not os.path.exists(self.sub_paths['vocabulary']):
            os.mkdir(self.sub_paths['vocabulary'])

        compile_pc1_uniques(self)
        compile_pc1_mappings_region(self)
        compile_global_uniques(self)

        path_thoroughfare_patterns = os.path.join(MODULE_PATH, 'parser', 'thoroughfare_patterns.json')
        shutil.copy(path_thoroughfare_patterns, self.sub_paths['thoroughfare_patterns'])

        print('Finished building vocabularies')

    def build_csv(self):

        for table_name in ['indexed', 'expanded']:

            print(f'Start saving {table_name} to csv for in-memory DB')

            header = True

            cols = self.get_columns_of_table(table_name)

            sql = f'select * from {table_name}'

            chunk_gen = pd.read_sql(sql, con=self.get_conn(), columns=cols, chunksize=int(1e5))

            for chunk in tqdm(chunk_gen):

                chunk.to_csv(os.path.join(self.sub_paths['csv'], f'{table_name}.csv.gz'),
                             header=header, mode='a', compression='gzip')
                header = False

    def get_conn(self):
        return sqlite3.connect(self.sub_paths['sql_db'], timeout=10)

    def get_conn_temp(self):
        return sqlite3.connect(self.sub_paths['sql_db_temp'], timeout=10)

    def delete_temp(self):
        try:
            os.remove(self.sub_paths['sql_db_temp'])
        except FileNotFoundError:
            pass

    def sql_query(self, query):

        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute(query)
        res = cur.fetchall()
        cur.close()
        conn.close()

        return res

    def sql_query_by_column_values(self, table_name, column, value_list):

        in_clause = ', '.join(value_list)
        res = self.sql_query(f"""select * from {table_name} where {column} IN ({in_clause})""")
        return res

    def sql_query_batch(self, query, batch_size=int(1e5)):
        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute(query)

        for batch in iter(lambda: cur.fetchmany(batch_size), []):
            yield batch

        conn.close()

    def sql_table_batch_by_column(self, table_name, by_columns, batch_size, match_func=None):

        if isinstance(by_columns, str):
            by_columns = [by_columns]

        table_columns = self.get_columns_of_table(table_name)

        column_indices = [table_columns.index(col) for col in by_columns]

        query = """
            SELECT *
            FROM %s
            ORDER BY %s
        """ % (table_name, ','.join(by_columns))

        con = self.get_conn()
        cur = con.cursor()
        cur.execute(query)

        res = cur.fetchmany(batch_size)
        next_res = cur.fetchone()

        while len(res):

            try:

                if match_func is None:
                    while all([next_res[i] == res[-1][i] for i in column_indices]):
                        res.append(next_res)
                        next_res = cur.fetchone()
                else:
                    while match_func(next_res, res):
                        res.append(next_res)
                        next_res = cur.fetchone()

            except TypeError:
                pass

            res_df = pd.DataFrame.from_records(res)
            res_df.columns = table_columns

            yield res_df

            if next_res is None:
                break
            else:
                res = [next_res]
                res.extend(cur.fetchmany(batch_size))
                next_res = cur.fetchone()

        con.close()

    def attach_uprn_fields_to_df(self, df, uprn_col='uprn'):

        uprn_list = [s for s in df[uprn_col].to_list() if isinstance(s, str)]

        res = self.sql_query_by_column_values(table_name='indexed', column='UPRN', value_list=uprn_list)
        columns = self.get_columns_of_table(table_name='indexed')
        res_df = pd.DataFrame.from_records(res, columns=columns)

        df = pd.merge(left=df, right=res_df, left_on=uprn_col, right_on='UPRN', how='left', suffixes=(None, '_db'))

        return df

    def create_index(self, table_name, columns):

        index_name = '__'.join(columns)

        sql = "CREATE INDEX IF NOT EXISTS %s ON %s(%s)" % (
            index_name,
            table_name,
            ', '.join(columns)
        )

        with self.get_conn() as conn:
            cur = conn.cursor()
            cur.execute(sql)



"""
Top level API for database setup and address matching

This module will be made available to module level
"""

from .database_setup import create_database
from .match import match
from .score import score
from .score_and_match import score_and_match